/*	motion.c
 *
 *	Detect changes in a video stream.
 *	Copyright 2000 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include "motion.h"

int deamon=0;
int snapshot=0;
int alarmtime=0;
int fmt=VIDEO_PALETTE_RGB24;
int read_img=0;

void usage (void)
{
	printf("motion Version "VERSION", Copyright 2000 Jeroen Vreeken\n");
	printf("\nusage:\tmotion [options]\n");
	printf("\n");
	printf("Overall settings:\n");
	printf("-C\t\tOutput changes count for every frame, usable for tuning\n");
	printf("-D\t\tDeamonize\n");
	printf("-a seconds\ttime between two automated snapshots, default: 0\n");
	printf("-c changes\tthreshold for detecting a change, default: %d\n", DEF_CHANGES);
	printf("-l\t\tLocate and mark motion on output pictures\n");
	printf("-m\t\tOutput 'motion' images\n");
	printf("-N\t\tDon't output normal images\n");
	printf("-t target dir\tdestination for snapshots.\n");
	printf("\nDevice settings:\n");
	printf("-d device\tvideo4linux capture device, default: "VIDEO_DEVICE"\n");
	printf("-i input\tinput channel to use, default: not set\n");
	printf("-n norm\t\tnorm to use (pal/ntsc/secam), default: pal\n");
	printf("-q quality\tJPEG image quality, default: %d\n", DEF_QUALITY);
	printf("-s widthxheight\tPicture size, default: %dx%d\n", DEF_WIDTH, DEF_HEIGHT);
	printf("\nActions:\n");
	printf("-E command\tExecute 'command' when detecting motion.\n");
	printf("-M address\tSend a mail to 'address' when detecting motion.\n");
	printf("-S nr\t\tSend a SMS to nr using sms_client when detecting motion.\n");
	printf("-g seconds\tminimum gap between events, default %d\n", DEF_GAP);
	printf("\n");
}

char *start_capture (int dev, int width, int height, int input, int norm)
{
        struct video_capability vid_caps;
	struct video_channel vid_chnl;
	struct video_window vid_win;
	struct video_mbuf vid_buf;
	char *map;

	if (ioctl (dev, VIDIOCGCAP, &vid_caps) == -1) {
		perror ("ioctl (VIDIOCGCAP)");
		return (NULL);
	}
	if (vid_caps.type & VID_TYPE_MONOCHROME) fmt=VIDEO_PALETTE_GREY;
	/* Set channels and norms, NOT TESTED my philips cam doesn't have a
	 * tuner. update: it works with my brothers haupage wintv (bttv driver)
	 */
        if (input != IN_DEFAULT) {
                vid_chnl.channel = -1;
                if (ioctl (dev, VIDIOCGCHAN, &vid_chnl) == -1) {
                        perror ("ioctl (VIDIOCGCHAN)");
                } else {
                        vid_chnl.channel = input;
                        vid_chnl.norm    = norm;
                        if (ioctl (dev, VIDIOCSCHAN, &vid_chnl) == -1) {
                                perror ("ioctl (VIDIOCSCHAN)");
                                return (NULL);
                        }
                }
        }
	if (ioctl (dev, VIDIOCGMBUF, &vid_buf) == -1) {
		fprintf(stderr, "no mmap falling back on read\n");
		if (ioctl (dev, VIDIOCGWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCGWIN");
			return (NULL);
		}
		vid_win.width=width;
		vid_win.height=height;
		if (ioctl (dev, VIDIOCSWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCSWIN");
			return (NULL);
		}
		read_img=1;
		map=malloc(width*height*3);
		return (map);
	}
	/* If we are going to capture greyscale we need room to blow the image up */
	if (fmt==VIDEO_PALETTE_GREY)
		map=mmap(0, vid_buf.size*3, PROT_READ|PROT_WRITE, MAP_SHARED, dev, 0);
	else
		map=mmap(0, vid_buf.size, PROT_READ|PROT_WRITE, MAP_SHARED, dev, 0);
	
	if ((unsigned char *)-1 == (unsigned char *)map)
		return (NULL);
	return map;
}

char *next_capture (int dev, char *map, int width, int height)
{
	int i;
	struct video_mmap vid_mmap;

	if (read_img) {
		if (fmt==VIDEO_PALETTE_GREY) {
			if (read(dev, map, width*height) <= 0)
				return NULL;
		} else {
			if (read(dev, map, width*height*3) <= 0)
				return NULL;
		}
	} else {
		vid_mmap.format=fmt;
		vid_mmap.frame=0;
		vid_mmap.width=width;
		vid_mmap.height=height;
		if (ioctl(dev, VIDIOCMCAPTURE, &vid_mmap) == -1)
			return (NULL);
		if (ioctl(dev, VIDIOCSYNC, &vid_mmap) == -1)
			return (NULL);
	}
	/* Blow up a greyscale image */
	if (fmt==VIDEO_PALETTE_GREY) {
		for (i=width*height*3; i>=0; i--) {
			map[i]=map[i/3];
		}
	}
	return map;
}

void put_image_jpeg (FILE *picture, char *image, int width, int height, int quality)
{
        int y, x, line_width;
        JSAMPROW row_ptr[1];
        struct jpeg_compress_struct cjpeg;
        struct jpeg_error_mgr jerr;
        unsigned char *line;

        line = malloc (width * 3);
        if (!line)
                return;
        cjpeg.err = jpeg_std_error(&jerr);
        jpeg_create_compress (&cjpeg);
        cjpeg.image_width = width;
        cjpeg.image_height= height;
        cjpeg.input_components = 3;
        cjpeg.in_color_space = JCS_RGB;
        jpeg_set_defaults (&cjpeg);

        jpeg_set_quality (&cjpeg, quality, TRUE);
        cjpeg.dct_method = JDCT_FASTEST;
        jpeg_stdio_dest (&cjpeg, picture);

        jpeg_start_compress (&cjpeg, TRUE);

        row_ptr[0] = line;
        line_width = width * 3;
	for ( y = 0; y < height; y++) {
        for (x = 0; x < line_width; x+=3) {
                        line[x]   = image[x+2];
                        line[x+1] = image[x+1];
                        line[x+2] = image[x];
                }
                jpeg_write_scanlines (&cjpeg, row_ptr, 1);
                image += line_width;
        }
        jpeg_finish_compress (&cjpeg);
        jpeg_destroy_compress (&cjpeg);
        free (line);
}

void send_sms(char *sms_nr, struct tm *currenttime)
{
	char message[150];

	if (!fork()) {
	    	printf("Sending SMS message to %s\n", sms_nr);
		sprintf(message, "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		execlp("sms_client", "sms_client", sms_nr, message, 0);
		perror("Unable to send SMS message");
		exit(0);
	}
	return;
}

void send_mail(char *mail_address, struct tm *currenttime)
{
	char message[150];
	int fdpipe[2];
	
	if (!fork()) {
	    	printf("Sending mail to %s\n", mail_address);
		sprintf(message, "This is an automated message generated by motion\n\n"
		    "Motion detected: %04d-%02d-%02d %02d:%02d:%02d\n"
		    ".\n",
		    currenttime->tm_year + 1900, currenttime->tm_mon +1,
		    currenttime->tm_mday, currenttime->tm_hour,
		    currenttime->tm_min, currenttime->tm_sec);
		pipe(fdpipe);
		if (!deamon) {
			close(0);
			dup(fdpipe[0]);
			close(fdpipe[0]);
		}
		write(fdpipe[1], message, strlen(message));
		close(fdpipe[1]);
		execlp("mail", "mail", mail_address, "-s", "MOTION ALERT", 0);
		perror("Unable to send message");
		exit(0);
	}
	return;
}

void exec_externcommand(char *externcommand)
{
	if (!fork()) {
		printf("Executing external command '%s'\n", externcommand);
		execlp(externcommand, externcommand, 0);
		perror("Unable to start external command");
		exit(-1);
	}
	return;
}

void locate_diffs(char *out, char *new, int width, int height)
{
	int x, y, maxx=0, maxy=0, minx=width, miny=height;

	/* Locate movement */	
	for (x=0; x<width; x++) {
		for (y=0; y<height; y++) {
			if (out[(x+y*width)*3+0] && out[(x+y*width)*3+1] && out[(x+y*width)*3+2]) {
				if (x < minx) minx=x;
				if (x > maxx) maxx=x;
				if (y < miny) miny=y;
				if (y > maxy) maxy=y;
			}
		}
	}
	/* Draw a box around the movement */
	for (x=minx; x<=maxx; x++) {
		new[(x+width*miny)*3+0]=255;
		new[(x+width*miny)*3+1]=255;
		new[(x+width*miny)*3+2]=255;
	}
	for (x=minx; x<=maxx; x++) {
		new[(x+width*maxy)*3+0]=255;
		new[(x+width*maxy)*3+1]=255;
		new[(x+width*maxy)*3+2]=255;
	}
	for (y=miny; y<=maxy; y++) {
		new[(minx+y*width)*3+0]=255;
		new[(minx+y*width)*3+1]=255;
		new[(minx+y*width)*3+2]=255;
	}
	for (y=miny; y<=maxy; y++) {
		new[(maxx+y*width)*3+0]=255;
		new[(maxx+y*width)*3+1]=255;
		new[(maxx+y*width)*3+2]=255;
	}
}

void sig_alarm (int signo) 
{
	/* Somebody (maybe we ourself) wants us to make a snapshot */
	snapshot=1;
	/* Set everything for the next snapshot */
	signal(SIGALRM, sig_alarm);
	return;
}

int main (int argc, char **argv)
{
	int width=DEF_WIDTH, height=DEF_HEIGHT, quality=DEF_QUALITY;
	int max_changes=DEF_CHANGES, always_diff=0, new_img=1, motion_img=0;
	int dev, i, diffs, shots=0, gap=DEF_GAP, locate=0, event_nr=0;
	int input=IN_DEFAULT, norm=0;
	unsigned char *image_ref, *image_new, *image_out;
	time_t currenttimep, lasttime=0, lastalarm=time(NULL);
	struct tm *currenttime=NULL;
	char *device=VIDEO_DEVICE, *mail_address=NULL, *sms_nr=NULL;
	char file[255], filebase[255], *filepath=NULL, *externcommand=NULL, c;
	FILE *picture;

	while ((c=getopt(argc, argv, "a:Cc:d:E:Dg:hi:lM:mNn:q:S:s:t:"))!=EOF) {
		switch (c) {
			case 'a':
				sscanf (optarg, "%d", &alarmtime);
				break;
			case 'C':
				always_diff=1;
				break;
			case 'c':
				sscanf (optarg, "%d", &max_changes);
				break;
			case 'd':
				device=optarg;
				break;
			case 'E':
				externcommand=optarg;
				break;
			case 'D':
				deamon=1;
				break;
			case 'g':
				sscanf (optarg, "%d", &gap);
				break;
			case 'i':
				sscanf (optarg, "%d", &input);
				break;
			case 'l':
				locate=1;
				break;
			case 'M':
				mail_address=optarg;
				break;
			case 'm':
				motion_img=1;
				break;
			case 'N':
				new_img=0;
				break;
			case 'n':
				if (!strcmp(optarg, "pal")) norm=NORM_PAL;
				if (!strcmp(optarg, "ntsc")) norm=NORM_NTSC;
				if (!strcmp(optarg, "secam")) norm=NORM_SECAM;
				break;
			case 'S':
				sms_nr=optarg;
				break;
			case 'q':
				sscanf (optarg, "%d", &quality);
				break;
			case 's':
				sscanf (optarg, "%dx%d", &width, &height);
				break;
			case 't':
				filepath=optarg;
				break;
			case 'h':
			/* Do the same as an invallid option, but without the
			 * 'invalid option' error.... */
			default:
				usage();
				exit(-1);
		}
	}
	/* Set signal handlers */
	signal(SIGCHLD, SIG_IGN);
	signal(SIGALRM, sig_alarm);

	/* Sanity check */
	if (!motion_img && !new_img) {
		new_img=1;
		printf("There has to be atleast one output format, -N ignored\n");
	}	

	image_ref=malloc(width*height*3);
	image_out=malloc(width*height*3);

	dev=open (device, O_RDWR);
	if (dev < 0) {
		perror ("Failed to open video device");
		exit(-1);
	}

	/* Deamonize */
	if (deamon) {
		close(0);
		close(1);
		close(2);
		if (fork()) exit(0);
	}
	/* set the device settings */
	image_new=start_capture (dev, width, height, input, norm);
	if (!image_new) {
		perror("Capture error");
		exit(-1);
	}
	/* Allow videodevice to settle in */
	sleep(1);
	/* Capture first image, or we will get an alarm on start */
	if (!next_capture(dev, image_new, width, height)) {
		perror("Capture error");
		exit(-1);
	}
	memcpy(image_ref, image_new, width*height*3);

	/* Should go on forever... unless you bought vaporware :) */
	while (next_capture(dev, image_new, width, height)) {
		/* Make a differences picture in image_out */
		diffs=0;
		for (i=0; i<width*height*3; i++) {
			image_out[i]=image_ref[i]-image_new[i];
			if ((signed char)image_out[i] > 32 || (signed char)image_out[i] < -32) {
				image_out[i]=image_new[i];
				diffs++;
			} else {
				image_out[i]=0;
			}
			/* Old image slowly decays, this will make it even harder on
			 * slow moving object to stay undetected
			 */
			image_ref[i]=(image_ref[i]+image_new[i])/2;
		}
		currenttimep=time(NULL);
		currenttime=localtime(&currenttimep);
		if (alarmtime && currenttimep - lastalarm >= alarmtime) {
			snapshot=1;
			lastalarm=currenttimep;
		}
		if (lasttime!=currenttimep) {
			shots=0;
		}
		/* The cat just walked in :) */
		if (diffs > max_changes) {
			printf("\007Motion detected %d\n", diffs);
			/* Take action if necesarry */
			if (currenttimep-lasttime>= gap) {
				if (sms_nr) send_sms(sms_nr, currenttime);
				if (mail_address) send_mail(mail_address, currenttime);
				if (externcommand) exec_externcommand(externcommand);
				event_nr++;
			}
			lasttime=currenttimep;
			shots++;
			fflush(0);
		} else if (always_diff) {
			printf("changes: %d\n", diffs);
		}
		if (diffs > max_changes || snapshot) {
			if (filepath) {
				sprintf(filebase, "%s/%d-%04d%02d%02d%02d%02d%02d-%02d",
				    filepath, event_nr,
				    currenttime->tm_year + 1900, currenttime->tm_mon + 1,
				    currenttime->tm_mday, currenttime->tm_hour,
				    currenttime->tm_min, currenttime->tm_sec, shots);
			} else {
				sprintf(filebase, "%d-%04d%02d%02d%02d%02d%02d-%02d",
				    event_nr,
				    currenttime->tm_year + 1900, currenttime->tm_mon + 1,
				    currenttime->tm_mday, currenttime->tm_hour,
				    currenttime->tm_min, currenttime->tm_sec, shots);
			}
			/* Output the latest picture 'image_new'.
			 * replace it with 'image_out' for neat effects, but less
			 * recognizable
			 */
			if (motion_img) {
				sprintf(file, "%sm.jpg", filebase);
				printf("Motion picture saved to: %s\n", file);
				picture=fopen(file, "w");
				if (locate && diffs > max_changes)
					locate_diffs(image_out, image_out, width, height);
				put_image_jpeg (picture, image_out, width, height, quality);
				fclose(picture);
			} 
			if (new_img) {
				sprintf(file, "%s.jpg", filebase);
				printf("Normal picture saved to: %s\n", file);
				picture=fopen(file, "w");
				if (locate && diffs > max_changes)
					locate_diffs(image_out, image_new, width, height);
				put_image_jpeg (picture, image_new, width, height, quality);
				fclose(picture);
			}
			snapshot=0;
		}
		/* Use this to minimize the number of pictures, also reduces
		 * the chance of a good one... However it can cause some
		 * seconds delay due to buffering in the v4l drivers...
		 */
//		sleep(1);
	}
	perror("Somebody stole the video device, lets hope we got his picture");
	close (dev);
	free(image_ref);
	free(image_out);
	exit(0);
}
